from .base import Service, ServiceMeta, SimpleServer
from .tcp import TCPServer
from .udp import UDPServer


__all__ = (
    'Service',
    'ServiceMeta',
    'SimpleServer',
    'TCPServer',
    'UDPServer',
)



import asyncio
import socket

from .base import SimpleServer
from ..utils import OptionsType, bind_socket


class TCPServer(SimpleServer):
    PROTO_NAME = 'tcp'

    def __init__(self, address: str = None, port: int = None,
                 options: OptionsType = (), sock=None):
        if not sock:
            if not (address and port):
                raise RuntimeError(
                    'You should pass socket instance or '
                    '"address" and "port" couple'
                )

            self.socket = bind_socket(
                address=address,
                port=port,
                options=options,
            )
        elif not isinstance(sock, socket.socket):
            raise ValueError('sock must be socket instance')
        else:
            self.socket = sock

        super().__init__()

    async def handle_client(self, reader: asyncio.StreamReader,
                            writer: asyncio.StreamWriter):
        raise NotImplementedError

    async def start(self):
        self.server = await asyncio.start_server(
            self.handle_client,
            sock=self.socket,
            loop=self.loop,
        )

    async def stop(self, exc: Exception = None):
        await super().stop(exc)
        await self.server.wait_closed()



import asyncio
import itertools
import logging.handlers
import socket
from multiprocessing import cpu_count
from typing import Any, Coroutine, Iterable, List, Tuple

import uvloop

from aiomisc.thread_pool import ThreadPoolExecutor


log = logging.getLogger(__name__)


def chunk_list(iterable: Iterable[Any], size: int):
    iterable = iter(iterable)

    item = list(itertools.islice(iterable, size))
    while item:
        yield item
        item = list(itertools.islice(iterable, size))


OptionsType = Iterable[Tuple[int, int, int]]


def bind_socket(*args, address: str, port: int, options: OptionsType = (),
                reuse_addr: bool = True, reuse_port: bool = False,
                proto_name: str = 'tcp'):

    if not args:
        if ':' in address:
            args = (socket.AF_INET6, socket.SOCK_STREAM)
        else:
            args = (socket.AF_INET, socket.SOCK_STREAM)

    sock = socket.socket(*args)
    sock.setblocking(0)

    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, int(reuse_addr))
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, int(reuse_port))

    for level, option, value in options:
        sock.setsockopt(level, option, value)

    sock.bind((address, port))
    sock_addr = sock.getsockname()[:2]

    if sock.family == socket.AF_INET6:
        log.info('Listening %s://[%s]:%s', proto_name, *sock_addr)
    else:
        log.info('Listening %s://%s:%s', proto_name, *sock_addr)

    return sock


def new_event_loop(pool_size=None) -> asyncio.AbstractEventLoop:
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    pool_size = pool_size or cpu_count()

    try:
        asyncio.get_event_loop().close()
    except RuntimeError:
        pass  # event loop is not created yet

    loop = asyncio.new_event_loop()
    thread_pool = ThreadPoolExecutor(pool_size, loop=loop)

    loop.set_default_executor(thread_pool)

    asyncio.set_event_loop(loop)

    return loop


# Type hint should be here for pylama's checks
_TASKS_LIST = List[asyncio.Task]


def wait_for(*coroutines: Tuple[Coroutine, ...],
             raise_first: bool = True,
             cancel: bool = True,
             loop: asyncio.AbstractEventLoop = None):

    tasks = list()                           # type: _TASKS_LIST
    loop = loop or asyncio.get_event_loop()  # type: asyncio.AbstractEventLoop
    result_future = loop.create_future()     # type: asyncio.Future
    waiting = len(coroutines)

    def cancel_pending():
        nonlocal result_future
        nonlocal tasks

        for t in tasks:
            if t.done():
                continue

            t.cancel()

    def raise_first_exception(exc: Exception):
        nonlocal result_future
        nonlocal tasks

        if result_future.done():
            return

        result_future.set_exception(exc)

    def return_result():
        nonlocal result_future
        nonlocal tasks

        if result_future.done():
            return

        results = []

        for task in tasks:
            exc = task.exception()

            results.append(task.result() if exc is None else exc)

        result_future.set_result(results)

    def done_callback(t: asyncio.Future):
        nonlocal tasks
        nonlocal result_future
        nonlocal waiting

        waiting -= 1

        cancelled = t.cancelled()

        exc = None
        if not cancelled:
            exc = t.exception()

        if cancelled or exc is None:
            if waiting == 0:
                return_result()

            return

        if raise_first:
            raise_first_exception(exc)

        if waiting == 0:
            return_result()

    for coroutine in coroutines:
        task = loop.create_task(coroutine)
        task.add_done_callback(done_callback)
        tasks.append(task)

    async def run():
        nonlocal result_future

        try:
            return await result_future
        finally:
            if cancel:
                cancel_pending()

    return run()



try:
    from .version import __version__, version_info
except ImportError:
    version_info = (0, 0, 0)
    __version__ = '{}.{}.{}'.format(*version_info)


authors = (
    ('Dmitry Orlov', 'me@mosquito.su'),
)

authors_email = ", ".join(
    '{}'.format(email) for _, email in authors
)

__license__ = 'MIT',
__author__ = ", ".join(
    '{} <{}>'.format(name, email) for name, email in authors
)

package_info = 'aiomisc - miscellaneous utils for asyncio'

# It's same persons right now
__maintainer__ = __author__

__all__ = (
    '__author__', '__author__', '__license__',
    '__maintainer__', '__version__',
    'version_info',
)



import asyncio
import logging
from contextlib import contextmanager
from typing import Tuple, Optional, Union

from .log import basic_config, LogFormat
from .service import Service
from .utils import new_event_loop


async def graceful_shutdown(services: Tuple[Service, ...],
                            loop: asyncio.AbstractEventLoop,
                            exception: Optional[Exception]):
    tasks = [
        asyncio.shield(loop.create_task(svc.stop(exception)), loop=loop)
        for svc in services
    ]

    if not tasks:
        return

    await asyncio.wait(tasks, loop=loop, return_when=asyncio.ALL_COMPLETED)


@contextmanager
def entrypoint(*services: Service,
               loop: asyncio.AbstractEventLoop = None,
               pool_size: int = None,
               log_level: Union[int, str] = logging.INFO,
               log_format: Union[str, LogFormat] = 'color',
               log_buffer_size: int = 1024,
               log_flush_interval: float = 0.2):

    loop = loop or new_event_loop(pool_size)

    basic_config(
        level=log_level,
        log_format=log_format,
        buffered=False,
    )

    async def start():
        nonlocal loop
        nonlocal services
        nonlocal log_format
        nonlocal log_level
        nonlocal log_buffer_size
        nonlocal log_flush_interval

        basic_config(
            level=log_level,
            log_format=log_format,
            buffered=True,
            loop=loop,
            buffer_size=log_buffer_size,
            flush_interval=log_flush_interval,
        )

        starting = []

        for svc in services:
            svc.set_loop(loop)

            starting.append(loop.create_task(svc.start()))

        await asyncio.gather(*starting, loop=loop)

    loop.run_until_complete(start())
    try:
        yield loop
    except Exception as e:
        loop.run_until_complete(graceful_shutdown(services, loop, e))
        raise
    else:
        loop.run_until_complete(graceful_shutdown(services, loop, None))
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())



import asyncio


class ServiceMeta(type):
    def __new__(cls, name, bases, namespace, **kwds):
        instance = type.__new__(cls, name, bases, dict(namespace))

        check_instance = all(
            asyncio.iscoroutinefunction(getattr(instance, method))
            for method in instance.__async_required__
        )

        if not check_instance:
            raise TypeError(
                'Following methods must be coroutine functions',
                tuple('%s.%s' % (name, m) for m in instance.__async_required__)
            )

        return instance


class Service(metaclass=ServiceMeta):
    __async_required__ = frozenset({'start', 'stop'})

    def __init__(self, **kwargs):
        self.loop = None
        self._set_params(**kwargs)

    def set_loop(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop

    def _set_params(self, **kwargs):
        for name, value in kwargs.items():
            setattr(self, name, value)

    async def start(self):
        raise NotImplementedError

    async def stop(self, exception: Exception = None):
        pass


class SimpleServer(Service):
    def __init__(self):
        self.server = None
        super().__init__()

    async def start(self):
        raise NotImplementedError

    async def stop(self, exc: Exception = None):
        self.server.close()



import os

from setuptools import setup, find_packages


from importlib.machinery import SourceFileLoader


module_name = 'aiomisc'

module = SourceFileLoader(
    module_name,
    os.path.join(module_name, '__init__.py')
).load_module()


def load_requirements(fname):
    """ load requirements from a pip requirements file """
    with open(fname) as f:
        line_iter = (line.strip() for line in f.readlines())
        return [line for line in line_iter if line and line[0] != '#']


setup(
    name=module_name.replace('_', '-'),
    version=module.__version__,
    author=module.__author__,
    author_email=module.authors_email,
    license=module.__license__,
    description=module.package_info,
    long_description=open("README.rst").read(),
    platforms="all",
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: Russian',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    packages=find_packages(exclude=['tests']),
    install_requires=load_requirements('requirements.txt'),
    extras_require={
        'develop': load_requirements('requirements.dev.txt'),
    },
)



import socket

from aiohttp.web import Application, AppRunner, SockSite
from aiohttp.helpers import AccessLogger

from .base import Service
from ..utils import bind_socket


class AIOHTTPService(Service):
    __async_required__ = frozenset({'start', 'create_application'})

    def __init__(self, address: str = None, port: int = None,
                 sock: socket.socket = None, shutdown_timeout: int = 5, **kwds):

        if not sock:
            if not (address and port):
                raise RuntimeError(
                    'You should pass socket instance or '
                    '"address" and "port" couple'
                )

            self.socket = bind_socket(
                address=address,
                port=port,
                proto_name='http',
            )

        elif not isinstance(sock, socket.socket):
            raise ValueError('sock must be socket instance')
        else:
            self.socket = sock

        self.runner = None
        self.site = None
        self.shutdown_timeout = shutdown_timeout

        super().__init__(**kwds)

    async def create_application(self) -> Application:
        raise NotImplementedError('You should implement '
                                  '"create_application" method')

    async def start(self):
        self.runner = AppRunner(
            await self.create_application(),
            access_log_class=AccessLogger,
            access_log_format=AccessLogger.LOG_FORMAT,
        )

        await self.runner.setup()

        self.site = SockSite(
            self.runner, self.socket,
            shutdown_timeout=self.shutdown_timeout
        )

        await self.site.start()

    async def stop(self, exception: Exception):
        try:
            await self.site.stop()
        finally:
            await self.runner.cleanup()
